/*
   Educational example: particle binning on Intel architectures.
   (c) Colfax International, 2015

   For more information, refer to the 3-part series of publications
   "Optimization Techniques for the Intel MIC Architecture" 
   at Colfax Research:
      http://colfaxresearch.com/?p=6
      http://colfaxresearch.com/?p=709
      http://colfaxresearch.com/?p=1231

*/

#include "Binning.h"
#include <cassert>
#include <string>


void BinParticles(const InputDataType  & inputData, BinsType & outputBins) {

  const int nThreads = omp_get_max_threads();

  // Instead of storing scalars in each bin,
  // store an array with values for each thread
  int globalBins[nThreads][nBinsX][nBinsY];

#pragma omp parallel
  {

    // Find this thread's own region in the global container 
    const int iThread = omp_get_thread_num();

    // Thread-private containers for partial bin counts
    for (int iX = 0; iX < nBinsX; iX++)
      for (int iY = 0; iY < nBinsY; iY++)
	globalBins[iThread][iX][iY] = 0;

    const int STRIP_WIDTH=16;

#pragma omp for
    for (int ii = 0; ii < inputData.numDataPoints; ii+=STRIP_WIDTH) { 

      // Temporary containers to carry indices over from one loop to the other
      int iX[STRIP_WIDTH] __attribute__((aligned(64)));
      int iY[STRIP_WIDTH] __attribute__((aligned(64)));

#pragma vector aligned
      for (int i = ii; i < ii+STRIP_WIDTH; i++) {

	// Transforming from cylindrical to Cartesian coordinates:
	const FTYPE x = inputData.r[i]*COS(inputData.phi[i]);
	const FTYPE y = inputData.r[i]*SIN(inputData.phi[i]);
      
	// Calculating the bin numbers for these coordinates:
	iX[i-ii] = int((x - xMin)*binsPerUnitX);
	iY[i-ii] = int((y - yMin)*binsPerUnitY);
      }
      
      for (int i = ii; i < ii+STRIP_WIDTH; i++) { 
	// Incrementing the appropriate bin in the counter,
	// protecting race condition with atomics
	globalBins[iThread][iX[i-ii]][iY[i-ii]]++;
      }

    }

  }

  // Reduction from global containers to the shared container
  for (int iThread = 0; iThread < nThreads; iThread++)
    for (int iX = 0; iX < nBinsX; iX++)
      for (int iY = 0; iY < nBinsY; iY++)
	outputBins[iX][iY] += globalBins[iThread][iX][iY];

}



int main(int argv, char* argc[]){

  // Greeting
  printf("\nParticle Binning Optimization Demo (%s)\n%s\n%s\n\n",
#ifdef DOUBLE_PRECISION
	 "double precision",
#else
	 "single precision",
#endif
	 "Additional information is available in accompanying papers at http://colfaxresearch.com/\n",
	 "(c) Colfax International, 2015.");

  // Allocate data
  t0 = omp_get_wtime(); printf("Initialization..."); fflush(stdout);
  InputDataType rawData; // Input of the calculation
  BinsType binnedData;   // Output of the calculation
  rawData.numDataPoints = n;
  rawData.r   = (FTYPE*) _mm_malloc(sizeof(FTYPE)*n, 64);
  rawData.phi = (FTYPE*) _mm_malloc(sizeof(FTYPE)*n, 64);

#pragma omp parallel for
  for (int i = 0; i < n; i++) {
    rawData.r[i] = 0.0;
    rawData.phi[i] = 0.0;
  }

  // Get reference results and performance
  BinsType binnedDataRef;
  const double refPerf = GetReferenceResults(rawData, binnedDataRef);

  // Run the benchmark of the optimized code
  printf("\nBenchmarking...\n\n"); fflush(stdout);
  printf("\033[1m%7s %10s %10s %10s %s\033[0m\n", "Trial", "Time, s", "Speedup", "GP/s", "*");

  const int nTrials = 10;
  const int skipTrials = 2;
  double perf = 0.0, dperf = 0.0;

  for (int t=1; t<=nTrials; t++){
    
    // Reset counters
    binnedData[:][:] = 0;

    // Compute and benchmark
    t0 = omp_get_wtime();
    BinParticles(rawData, binnedData);
    t1 = omp_get_wtime();

    // Validating results
    if (t == 1) ValidateResults(binnedData, binnedDataRef);
      
    // Translate wall clock time to performance units
    if (t > skipTrials) {
      // Ignore the first few trials: those are warm-up
      perf += HztoPerf/(t1-t0);
      dperf += HztoPerf*HztoPerf/((t1-t0)*(t1-t0));
    }
      
    printf("%7d %10.3e %10.2f %10.2e %s\n", 
	   t, (t1-t0), HztoPerf/((t1-t0)*refPerf), HztoPerf/(t1-t0), (t<=skipTrials?"**":""));
    
    fflush(stdout);
   
  }
   
  perf/=(double)(nTrials-skipTrials); 
  dperf=sqrt(dperf/(double)(nTrials-skipTrials)-perf*perf);
  printf("---------------------------------------------------------\n");
  printf("\033[1m%s%7.2f \033[42m%10.2e +- %.2e GP/s\033[0m\n",
	 "Optimized performance:", perf/refPerf, perf, dperf);
  printf("---------------------------------------------------------\n");
  printf("*  - Performance unit 1 GP/s is 10^9 particles binned per second.\n"); 
  printf("** - warm-up, not included in average\n\n"); fflush(stdout);

  _mm_free(rawData.r);
  _mm_free(rawData.phi);
}
